@@ -18,6 +18,8 @@ import os, re, sys, time, weakref
 from collections import deque
 from functools import partial
 from threading import Lock
+from subprocess import Popen, PIPE
+import cStringIO
 #from traceback import print_stack
 
 __all__ = ["git_version", "GitError", "GitErrorSha", "Storage", "StorageFactory"]
@@ -33,19 +35,29 @@ class GitCore:
         self.__git_bin = git_bin
         self.__git_dir = git_dir
 
-    def __execute2(self, gitcmd, *args):
-        # construct command tuple
+    def __build_git_cmd(self, gitcmd, *args):
+        "construct command tuple for git call suitable for Popen()"
+
         cmd = [self.__git_bin]
         if self.__git_dir:
             cmd.append('--git-dir=%s' % self.__git_dir)
         cmd.append(gitcmd)
         cmd.extend(args)
 
-        #print >>sys.stderr, "GitCore '%s'" % str(cmd)
-        return os.popen3(cmd) # (input, output, error)
+        return cmd
 
     def __execute(self, git_cmd, *cmd_args):
-        return self.__execute2(git_cmd, *cmd_args)[1]
+        "execute git command and return file-like object of stdout"
+
+        #print >>sys.stderr, "DEBUG:", git_cmd, cmd_args
+
+        p = Popen(self.__build_git_cmd(git_cmd, *cmd_args),
+                  stdin=None, stdout=PIPE, stderr=PIPE, close_fds=True)
+
+        stdout_data, stderr_data = p.communicate()
+        #TODO, do something with p.returncode, e.g. raise exception
+
+        return cStringIO.StringIO(stdout_data)
 
     def __getattr__(self, name):
         return partial(self.__execute, name.replace('_','-'))
@@ -228,7 +240,7 @@ class Storage:
                 new_tags = set([])
                 youngest = None
                 oldest = None
-                for revs in self.repo.rev_parse("--tags").readlines():
+                for revs in self.repo.rev_parse("--tags"):
                     new_tags.add(revs.strip())
 
                 # helper for reusing strings
@@ -238,7 +250,7 @@ class Storage:
                     return __rev_seen.setdefault(rev, rev)
 
                 rev = ord_rev = 0
-                for revs in self.repo.rev_list("--parents", "--all").readlines():
+                for revs in self.repo.rev_list("--parents", "--all"):
                     revs = revs.strip().split()
 
                     revs = map(__rev_reuse, revs)
@@ -313,7 +325,7 @@ class Storage:
                 self.__rev_cache = youngest, oldest, new_db, new_tags, new_sdb
                 self.logger.debug("rebuilt commit tree db for %d with %d entries" % (id(self),len(new_db)))
 
-            assert all(e is not None for e in self.__rev_cache)
+            assert all(e is not None for e in self.__rev_cache) or not any(self.__rev_cache)
 
             return self.__rev_cache
         # with self.__rev_cache_lock
@@ -454,7 +466,7 @@ class Storage:
     def get_branches(self):
         "returns list of (local) branches, with active (= HEAD) one being the first item"
         result=[]
-        for e in self.repo.branch("-v", "--no-abbrev").readlines():
+        for e in self.repo.branch("-v", "--no-abbrev"):
             (bname,bsha)=e[1:].strip().split()[:2]
             if e.startswith('*'):
                 result.insert(0,(bname,bsha))
@@ -463,18 +475,29 @@ class Storage:
         return result
 
     def get_tags(self):
-        return [e.strip() for e in self.repo.tag("-l").readlines()]
+        return [e.strip() for e in self.repo.tag("-l")]
 
     def ls_tree(self, rev, path=""):
         rev = str(rev) # paranoia
         if path.startswith('/'):
             path = path[1:]
-        return [e.split(None, 3) for e in \
-                    self.repo.ls_tree("-z", rev, "--", path).read().split('\0') if e]
+
+        if path:
+            tree = self.repo.ls_tree("-z", rev, "--", path)
+        else:
+            tree = self.repo.ls_tree("-z", rev)
+
+        def split_ls_tree_line(l):
+            "split according to '<mode> <type> <sha>\t<fname>'"
+            meta,fname = l.split('\t')
+            _mode,_type,_sha = meta.split(' ')
+            return _mode,_type,_sha,fname
+
+        return [split_ls_tree_line(e) for e in tree.read().split('\0') if e]
 
     def read_commit(self, commit_id):
         if not commit_id:
-            raise GitErrorCommit_Id
+            raise GitError("read_commit called with empty commit_id")
 
         commit_id = str(commit_id)
 
@@ -578,14 +601,14 @@ class Storage:
         if limit is None:
             limit = -1
         for rev in self.repo.rev_list("--max-count=%d" % limit,
-                                      str(sha), "--", path).readlines():
+                                      str(sha), "--", path):
             yield rev.strip()
 
     def history_timerange(self, start, stop):
         for rev in self.repo.rev_list("--reverse",
                                       "--max-age=%d" % start,
                                       "--min-age=%d" % stop,
-                                      "--all").readlines():
+                                      "--all"):
             yield rev.strip()
 
     def rev_is_anchestor_of(self, rev1, rev2):
@@ -597,7 +620,7 @@ class Storage:
     def blame(self, commit_sha, path):
         in_metadata = False
 
-        for line in self.repo.blame("-p", "--", path, str(commit_sha)).readlines():
+        for line in self.repo.blame("-p", "--", path, str(commit_sha)):
             assert line
             if in_metadata:
                 in_metadata = not line.startswith('\t')
@@ -633,7 +656,7 @@ class Storage:
         assert lines[-1] == ""
         del lines[-1]
 
-        if tree1 is None:
+        if tree1 is None and lines:
             # if only one tree-sha is given on commandline,
             # the first line is just the redundant tree-sha itself...
             assert not lines[0].startswith(':')
